from abc import abstractmethod

import torch
from torch import Tensor, nn
from torch.distributions import Distribution
from torch.optim import Optimizer

from rl_replicas.common.policies.policy import Policy


class StochasticPolicy(Policy):
    """
    The abstract base class for stochastic policies.

    :param network: (nn.Module) The network.
    :param optimizer: (Optimizer) The optimizer.
    """

    def __init__(self, network: nn.Module, optimizer: Optimizer):
        super().__init__(network, optimizer)

    @abstractmethod
    def forward(self, observation: Tensor) -> Distribution:
        """
        Forward pass in policy

        :param observation: (Tensor) The observation of the environment
        :return: (Distribution) The distribution of action(s).
        """
        raise NotImplementedError

    def predict(self, observation: Tensor) -> Tensor:
        """
        Selects the action(s) based on the observation of the environment.

        :param observation: (Tensor) The observation of the environment
        :return: (Tensor) the action(s)
        """
        with torch.no_grad():
            distribution: Distribution = self.forward(observation)

        action: Tensor = distribution.sample()

        return action



"""
Policies
"""
from rl_replicas.common.policies.categorical_policy import CategoricalPolicy
from rl_replicas.common.policies.deterministic_policy import DeterministicPolicy
from rl_replicas.common.policies.gaussian_policy import GaussianPolicy
from rl_replicas.common.policies.policy import Policy
from rl_replicas.common.policies.stochastic_policy import StochasticPolicy



import copy
import logging
from typing import Optional

import gym
import torch
from torch import Tensor
from torch.distributions import Distribution
from torch.nn import functional as F

from rl_replicas.common.base_algorithms.on_policy_algorithm import (
    OneEpochExperience,
    OnPolicyAlgorithm,
)
from rl_replicas.common.policies import Policy
from rl_replicas.common.value_function import ValueFunction

logger = logging.getLogger(__name__)


class PPO(OnPolicyAlgorithm):
    """
    Proximal Policy Optimization (by clipping) with early stopping based on approximate KL divergence

    :param policy: (Policy) The policy
    :param value_function: (ValueFunction) The value function
    :param env: (gym.Env) The environment to learn from
    :param clip_range: (float) The limit on the likelihood ratio between policies for clipping in the policy objective.
    :param max_kl_divergence: (float) The limit on the KL divergence between policies for early stopping.
    :param gamma: (float) The discount factor for the cumulative return
    :param gae_lambda: (float) The factor for trade-off of bias vs variance for Generalized Advantage Estimator
    :param seed: (int) The seed for the pseudo-random generators
    :param n_value_gradients (int): Number of gradient descent steps to take on value function per epoch.
    """

    def __init__(
        self,
        policy: Policy,
        value_function: ValueFunction,
        env: gym.Env,
        clip_range: float = 0.2,
        max_kl_divergence: float = 0.01,
        gamma: float = 0.99,
        gae_lambda: float = 0.97,
        seed: Optional[int] = None,
        n_policy_gradients: int = 80,
        n_value_gradients: int = 80,
    ) -> None:
        super().__init__(
            policy=policy,
            value_function=value_function,
            env=env,
            gamma=gamma,
            gae_lambda=gae_lambda,
            seed=seed,
            n_value_gradients=n_value_gradients,
        )

        self.clip_range = clip_range
        self.n_policy_gradients = n_policy_gradients
        self.max_kl_divergence = max_kl_divergence

        self.old_policy: Policy = copy.deepcopy(self.policy)

    def train(
        self,
        one_epoch_experience: OneEpochExperience,
    ) -> None:
        observations: Tensor = one_epoch_experience["observations"]
        actions: Tensor = one_epoch_experience["actions"]
        advantages: Tensor = one_epoch_experience["advantages"]

        # Normalize advantage
        advantages = (advantages - torch.mean(advantages)) / torch.std(advantages)

        # for logging
        with torch.no_grad():
            policy_dist: Distribution = self.policy(observations)
            policy_loss_before: Tensor = self.compute_policy_loss(
                observations, actions, advantages
            ).detach()
            log_probs: Tensor = policy_dist.log_prob(actions)
            entropies: Tensor = policy_dist.entropy()

        # Train policy
        for i in range(self.n_policy_gradients):
            policy_loss: Tensor = self.compute_policy_loss(
                observations, actions, advantages
            )
            approximate_kl_divergence: Tensor = self.compute_approximate_kl_divergence(
                observations, actions
            ).detach()
            if approximate_kl_divergence > 1.5 * self.max_kl_divergence:
                logger.info(
                    "Early stopping at update {} due to reaching max KL divergence.".format(
                        i
                    )
                )
                break

            self.policy.optimizer.zero_grad()
            policy_loss.backward()
            self.policy.optimizer.step()

        self.old_policy.load_state_dict(self.policy.state_dict())

        discounted_returns: Tensor = one_epoch_experience["discounted_returns"]

        # for logging
        with torch.no_grad():
            value_loss_before: Tensor = self.compute_value_loss(
                observations, discounted_returns
            )

        # Train value function
        for _ in range(self.n_value_gradients):
            value_loss: Tensor = self.compute_value_loss(
                observations, discounted_returns
            )
            self.value_function.optimizer.zero_grad()
            value_loss.backward()
            self.value_function.optimizer.step()

        logger.info("Policy Loss:            {:<8.3g}".format(policy_loss_before))
        logger.info("Avarage Entropy:        {:<8.3g}".format(torch.mean(entropies)))
        logger.info("Log Prob STD:           {:<8.3g}".format(torch.std(log_probs)))
        logger.info(
            "KL divergence:          {:<8.3g}".format(approximate_kl_divergence)
        )

        logger.info("Value Function Loss:    {:<8.3g}".format(value_loss_before))

        if self.tensorboard:
            self.writer.add_scalar(
                "policy/loss",
                policy_loss_before,
                self.current_total_steps,
            )
            self.writer.add_scalar(
                "policy/avarage_entropy",
                torch.mean(entropies),
                self.current_total_steps,
            )
            self.writer.add_scalar(
                "policy/log_prob_std",
                torch.std(log_probs),
                self.current_total_steps,
            )
            self.writer.add_scalar(
                "policy/approximate_kl_divergence",
                approximate_kl_divergence,
                self.current_total_steps,
            )

            self.writer.add_scalar(
                "value/loss",
                value_loss_before,
                self.current_total_steps,
            )

    def compute_policy_loss(
        self,
        observations: Tensor,
        actions: Tensor,
        advantages: Tensor,
    ) -> Tensor:
        policy_dist: Distribution = self.policy(observations)
        log_probs: Tensor = policy_dist.log_prob(actions)

        with torch.no_grad():
            old_policy_dist: Distribution = self.old_policy(observations)
            old_log_probs: Tensor = old_policy_dist.log_prob(actions)

        # Calculate surrogate
        likelihood_ratio: Tensor = torch.exp(log_probs - old_log_probs)
        surrogate: Tensor = likelihood_ratio * advantages

        # Clipping the constraint
        likelihood_ratio_clip: Tensor = torch.clamp(
            likelihood_ratio,
            min=1 - self.clip_range,
            max=1 + self.clip_range,
        )

        # Calculate surrotate clip
        surrogate_clip: Tensor = likelihood_ratio_clip * advantages

        policy_loss: Tensor = -torch.min(surrogate, surrogate_clip).mean()

        return policy_loss

    def compute_approximate_kl_divergence(
        self,
        observations: Tensor,
        actions: Tensor,
    ) -> Tensor:
        with torch.no_grad():
            policy_dist: Distribution = self.policy(observations)
            log_probs: Tensor = policy_dist.log_prob(actions)

            old_policy_dist: Distribution = self.old_policy(observations)
            old_log_probs: Tensor = old_policy_dist.log_prob(actions)

        approximate_kl_divergence: Tensor = old_log_probs - log_probs

        return torch.mean(approximate_kl_divergence)

    def compute_value_loss(
        self,
        observations: Tensor,
        discounted_returns: Tensor,
    ) -> Tensor:
        values: Tensor = self.value_function(observations)
        squeezed_values: Tensor = torch.squeeze(values, -1)
        value_loss: Tensor = F.mse_loss(squeezed_values, discounted_returns)

        return value_loss



import random
from operator import itemgetter
from typing import Dict, List

import torch
from torch import Tensor


class ReplayBuffer:
    """
    Replay buffer for off-policy algorithms.

    :param buffer_size: (int) Max number of element in the buffer
    """

    def __init__(self, buffer_size: int = int(1e6)) -> None:
        self.buffer_size = buffer_size

        self.current_size: int = 0
        self.observations: List[Tensor] = []
        self.actions: List[Tensor] = []
        self.rewards: List[float] = []
        self.next_observations: List[Tensor] = []
        self.dones: List[bool] = []

    def add_one_epoch_experience(
        self,
        observations: List[Tensor],
        actions: List[Tensor],
        rewards: List[float],
        next_observations: List[Tensor],
        dones: List[bool],
    ) -> None:
        self.observations.extend(observations)
        self.actions.extend(actions)
        self.rewards.extend(rewards)
        self.next_observations.extend(next_observations)
        self.dones.extend(dones)

        self.current_size += len(observations)

        if self.current_size > self.buffer_size:
            num_exceeded_experinece: int = self.current_size - self.buffer_size

            for _ in range(num_exceeded_experinece):
                self.observations.pop()
                self.actions.pop()
                self.rewards.pop()
                self.next_observations.pop()
                self.dones.pop()

                self.current_size -= 1

    def sample_minibatch(self, minibatch_size: int = 32) -> Dict[str, Tensor]:
        indices = random.sample(range(0, self.current_size), minibatch_size)

        sampled_observations: Tensor = torch.stack(
            itemgetter(*indices)(self.observations)
        )
        sampled_actions: Tensor = torch.stack(itemgetter(*indices)(self.actions))
        sampled_rewards: Tensor = Tensor(itemgetter(*indices)(self.rewards))
        sampled_next_observations: Tensor = torch.stack(
            itemgetter(*indices)(self.next_observations)
        )
        sampled_dones: Tensor = Tensor(itemgetter(*indices)(self.dones))

        minibatch: Dict[str, Tensor] = {
            "observations": sampled_observations,
            "actions": sampled_actions,
            "rewards": sampled_rewards,
            "next_observations": sampled_next_observations,
            "dones": sampled_dones,
        }

        return minibatch



import copy
import logging
from typing import Dict, List, Optional

import gym
import numpy as np
import torch
from torch import Tensor
from torch.nn import functional as F

from rl_replicas.common.base_algorithms import OffPolicyAlgorithm
from rl_replicas.common.policies import Policy
from rl_replicas.common.q_function import QFunction
from rl_replicas.common.replay_buffer import ReplayBuffer
from rl_replicas.common.utils import polyak_average

logger = logging.getLogger(__name__)


class TD3(OffPolicyAlgorithm):
    """
    Twin Delayed Deep Deterministic Policy Gradient (TD3)

    :param policy: (Policy) The policy
    :param q_function_1: (QFunction) The Q function
    :param q_function_2: (QFunction) The Q function
    :param env: (gym.Env) The environment to learn from
    :param gamma: (float) The discount factor for the cumulative return
    :param tau: (float) The interpolation factor in polyak averaging for target networks
    :param action_noise_scale: (float) The scale of the noise (std) for the policy to explore better
    :param target_noise_scale: (float) The scale of the smoothing noise (std) for the target policy to exploit harder
    :param target_noise_clip: (float) The limit for absolute value of the target policy smoothing noise
    :param policy_delay: (int) The policy will only be updated once every policy_delay times for each update of
        the Q-networks.
    :param seed: (int) The seed for the pseudo-random generators
    """

    def __init__(
        self,
        policy: Policy,
        q_function_1: QFunction,
        q_function_2: QFunction,
        env: gym.Env,
        gamma: float = 0.99,
        tau: float = 0.005,
        action_noise_scale: float = 0.1,
        target_noise_scale: float = 0.2,
        target_noise_clip: float = 0.5,
        policy_delay: int = 2,
        seed: Optional[int] = None,
    ) -> None:
        super().__init__(
            policy=policy,
            q_function=q_function_1,
            env=env,
            gamma=gamma,
            tau=tau,
            action_noise_scale=action_noise_scale,
            seed=seed,
        )

        self.target_noise_scale = target_noise_scale
        self.target_noise_clip = target_noise_clip
        self.policy_delay = policy_delay

        self.q_function_1 = self.q_function
        self.q_function_2 = q_function_2
        self.target_q_function_1 = self.target_q_function
        self.target_q_function_2 = copy.deepcopy(q_function_2)

        for param in self.target_q_function_2.network.parameters():
            param.requires_grad = False

    def train(
        self, replay_buffer: ReplayBuffer, train_steps: int, minibatch_size: int
    ) -> None:
        policy_losses: List[float] = []
        q_function_1_losses: List[float] = []
        q_function_2_losses: List[float] = []
        all_q_values_1: List[float] = []
        all_q_values_2: List[float] = []

        for train_step in range(train_steps):
            minibatch: Dict[str, Tensor] = replay_buffer.sample_minibatch(
                minibatch_size
            )

            observations: Tensor = minibatch["observations"]
            actions: Tensor = minibatch["actions"]
            rewards: Tensor = minibatch["rewards"]
            next_observations: Tensor = minibatch["next_observations"]
            dones: Tensor = minibatch["dones"]

            q_values_1: Tensor = self.q_function_1(observations, actions)
            q_values_2: Tensor = self.q_function_2(observations, actions)
            all_q_values_1.extend(q_values_1.tolist())
            all_q_values_2.extend(q_values_2.tolist())

            with torch.no_grad():
                next_actions: Tensor = self.target_policy(next_observations)
                epsilon: Tensor = self.target_noise_scale * torch.randn_like(
                    next_actions
                )
                epsilon = torch.clamp(
                    epsilon, -self.target_noise_clip, self.target_noise_clip
                )
                next_actions = next_actions + epsilon
                next_actions = torch.clamp(
                    next_actions, -self.action_limit, self.action_limit
                )

                target_q_values_1: Tensor = self.target_q_function_1(
                    next_observations, next_actions
                )
                target_q_values_2: Tensor = self.target_q_function_2(
                    next_observations, next_actions
                )
                target_q_values: Tensor = torch.min(
                    target_q_values_1, target_q_values_2
                )

                targets: Tensor = rewards + self.gamma * (1 - dones) * target_q_values

            q_function_1_loss: Tensor = F.mse_loss(q_values_1, targets)
            q_function_2_loss: Tensor = F.mse_loss(q_values_2, targets)
            q_function_1_losses.append(q_function_1_loss.item())
            q_function_2_losses.append(q_function_2_loss.item())

            self.q_function_1.optimizer.zero_grad()
            q_function_1_loss.backward()
            self.q_function_1.optimizer.step()

            self.q_function_2.optimizer.zero_grad()
            q_function_2_loss.backward()
            self.q_function_2.optimizer.step()

            if train_step % self.policy_delay == 0:
                # Freeze Q-networks
                for param in self.q_function_1.network.parameters():
                    param.requires_grad = False
                for param in self.q_function_2.network.parameters():
                    param.requires_grad = False

                policy_actions: Tensor = self.policy(observations)
                policy_q_values: Tensor = self.q_function_1(
                    observations, policy_actions
                )

                policy_loss: Tensor = -torch.mean(policy_q_values)
                policy_losses.append(policy_loss.item())

                self.policy.optimizer.zero_grad()
                policy_loss.backward()
                self.policy.optimizer.step()

                # Unfreeze Q-networks
                for param in self.q_function_1.network.parameters():
                    param.requires_grad = True
                for param in self.q_function_2.network.parameters():
                    param.requires_grad = True

                polyak_average(
                    self.policy.network.parameters(),
                    self.target_policy.network.parameters(),
                    self.tau,
                )
                polyak_average(
                    self.q_function_1.network.parameters(),
                    self.target_q_function_1.network.parameters(),
                    self.tau,
                )
                polyak_average(
                    self.q_function_2.network.parameters(),
                    self.target_q_function_2.network.parameters(),
                    self.tau,
                )

                if self.tensorboard:
                    self.writer.add_scalar(
                        "policy/loss", policy_loss, self.current_total_steps
                    )

            if self.tensorboard:
                self.writer.add_scalar(
                    "q-function_1/loss", q_function_1_loss, self.current_total_steps
                )
                self.writer.add_scalar(
                    "q-function_2/loss", q_function_2_loss, self.current_total_steps
                )
                self.writer.add_scalar(
                    "q-function_1/avarage_q-value",
                    torch.mean(q_values_1),
                    self.current_total_steps,
                )
                self.writer.add_scalar(
                    "q-function_2/avarage_q-value",
                    torch.mean(q_values_2),
                    self.current_total_steps,
                )

        logger.info("Policy Loss:            {:<8.3g}".format(np.mean(policy_losses)))
        logger.info(
            "Q Function Loss (1):    {:<8.3g}".format(np.mean(q_function_1_losses))
        )
        logger.info(
            "Q Function Loss (2):    {:<8.3g}".format(np.mean(q_function_2_losses))
        )

        logger.info("Average Q Value (1):    {:<8.3g}".format(np.mean(all_q_values_1)))
        logger.info("Max Q Value (1):        {:<8.3g}".format(np.max(all_q_values_1)))
        logger.info("Min Q Value (1):        {:<8.3g}".format(np.min(all_q_values_1)))

        logger.info("Average Q Value (2):    {:<8.3g}".format(np.mean(all_q_values_2)))
        logger.info("Max Q Value (2):        {:<8.3g}".format(np.max(all_q_values_2)))
        logger.info("Min Q Value (2):        {:<8.3g}".format(np.min(all_q_values_2)))

    def save_model(self, current_epoch: int, model_path: str) -> None:
        """
        Save model

        :param current_epoch: (int) The current epoch
        :param model_path: (int) The path to save the model
        """
        torch.save(
            {
                "epoch": current_epoch,
                "total_steps": self.current_total_steps,
                "policy_state_dict": self.policy.network.state_dict(),
                "policy_optimizer_state_dict": self.policy.optimizer.state_dict(),
                "target_policy_state_dict": self.target_policy.network.state_dict(),
                "q_function_1_state_dict": self.q_function_1.network.state_dict(),
                "q_function_1_optimizer_state_dict": self.q_function_1.optimizer.state_dict(),
                "target_q_function_1_state_dict": self.target_q_function_1.network.state_dict(),
                "q_function_2_state_dict": self.q_function_2.network.state_dict(),
                "q_function_2_optimizer_state_dict": self.q_function_2.optimizer.state_dict(),
                "target_q_function_2_state_dict": self.target_q_function_2.network.state_dict(),
            },
            model_path,
        )



import logging
import os
import time
from abc import ABC, abstractmethod
from typing import List, Optional

import gym
import numpy as np
import torch
from torch import Tensor
from torch.distributions import Distribution
from torch.utils.tensorboard import SummaryWriter
from typing_extensions import TypedDict

from rl_replicas.common.policies import Policy
from rl_replicas.common.utils import (
    discount_cumulative_sum,
    gae,
    seed_random_generators,
)
from rl_replicas.common.value_function import ValueFunction

logger = logging.getLogger(__name__)


class OneEpochExperience(TypedDict):
    observations: Tensor
    actions: Tensor
    advantages: Tensor
    discounted_returns: Tensor
    episode_returns: List[float]
    episode_lengths: List[int]


class OnPolicyAlgorithm(ABC):
    """
    The base of on-policy algorithms

    :param policy: (Policy) The policy
    :param value_function: (ValueFunction) The value function
    :param env: (gym.Env) The environment to learn from
    :param gamma: (float) The discount factor for the cumulative return
    :param gae_lambda: (float) The factor for trade-off of bias vs variance for Generalized Advantage Estimator
    :param seed: (int) The seed for the pseudo-random generators
    :param n_value_gradients (int): Number of gradient descent steps to take on value function per epoch.
    """

    def __init__(
        self,
        policy: Policy,
        value_function: ValueFunction,
        env: gym.Env,
        gamma: float,
        gae_lambda: float,
        seed: Optional[int],
        n_value_gradients: int,
    ) -> None:
        self.policy = policy
        self.value_function = value_function
        self.env = env
        self.gamma = gamma
        self.gae_lambda = gae_lambda
        if seed is not None:
            self.seed: int = seed
        self.n_value_gradients = n_value_gradients

        if seed is not None:
            self._seed()

    def _seed(self) -> None:
        seed_random_generators(self.seed)
        self.env.action_space.seed(self.seed)
        self.env.seed(self.seed)

    def learn(
        self,
        epochs: int = 50,
        steps_per_epoch: int = 4000,
        output_dir: str = ".",
        tensorboard: bool = False,
        model_saving: bool = False,
    ) -> None:
        """
        Learn the model

        :param epochs: (int) The number of epochs to run and train.
        :param steps_per_epoch: (int) The number of steps to run per epoch.
        :param output_dir: (str) The directory of output
        :param tensorboard: (bool) Whether or not to log for tensorboard
        :param model_saving: (bool) Whether or not to save trained model (Save and overwrite at each end of epoch)
        """
        self.tensorboard = tensorboard

        if self.tensorboard:
            logger.info("Set up tensorboard")
            os.makedirs(output_dir, exist_ok=True)
            tensorboard_path: str = os.path.join(output_dir, "tensorboard")
            self.writer: SummaryWriter = SummaryWriter(tensorboard_path)

        start_time: float = time.time()
        self.current_total_steps: int = 0
        self.current_total_episodes: int = 0

        for current_epoch in range(epochs):

            one_epoch_experience: OneEpochExperience = (
                self.collect_one_epoch_experience(steps_per_epoch)
            )

            if model_saving:
                logger.info("Set up model saving")
                os.makedirs(output_dir, exist_ok=True)
                model_path: str = os.path.join(output_dir, "model.pt")

                logger.info("Save model")
                torch.save(
                    {
                        "epoch": current_epoch,
                        "total_steps": self.current_total_steps,
                        "policy_state_dict": self.policy.network.state_dict(),
                        "policy_optimizer_state_dict": self.policy.optimizer.state_dict(),
                        "value_fn_state_dict": self.value_function.network.state_dict(),
                        "value_fn_optimizer_state_dict": self.value_function.optimizer.state_dict(),
                    },
                    model_path,
                )

            episode_returns: List[float] = one_epoch_experience["episode_returns"]
            episode_lengths: List[int] = one_epoch_experience["episode_lengths"]

            logger.info("Epoch: {}".format(current_epoch))

            logger.info(
                "Total steps:            {:<8.3g}".format(self.current_total_steps)
            )
            logger.info(
                "Total episodes:         {:<8.3g}".format(self.current_total_episodes)
            )

            logger.info(
                "Average Episode Return: {:<8.3g}".format(np.mean(episode_returns))
            )
            logger.info(
                "Episode Return STD:     {:<8.3g}".format(np.std(episode_returns))
            )
            logger.info(
                "Max Episode Return:     {:<8.3g}".format(np.max(episode_returns))
            )
            logger.info(
                "Min Episode Return:     {:<8.3g}".format(np.min(episode_returns))
            )

            logger.info(
                "Average Episode Length: {:<8.3g}".format(np.mean(episode_lengths))
            )

            logger.info(
                "Time:                   {:<8.3g}".format(time.time() - start_time)
            )

            if self.tensorboard:
                self.writer.add_scalar(
                    "training/average_episode_return",
                    np.mean(episode_returns),
                    self.current_total_steps,
                )
                self.writer.add_scalar(
                    "training/average_episode_length",
                    np.mean(episode_lengths),
                    self.current_total_steps,
                )

            self.train(one_epoch_experience)

        if self.tensorboard:
            self.writer.flush()
            self.writer.close()

    def collect_one_epoch_experience(self, steps_per_epoch: int) -> OneEpochExperience:
        one_epoch_experience: OneEpochExperience = {
            "observations": Tensor(),
            "actions": Tensor(),
            "advantages": Tensor(),
            "discounted_returns": Tensor(),
            "episode_returns": [],
            "episode_lengths": [],
        }

        observations_list: List[Tensor] = []
        actions_list: List[Tensor] = []

        advantages_ndarray: np.ndarray = np.zeros(steps_per_epoch, dtype=np.float32)
        discounted_returns_ndarray: np.ndarray = np.zeros(
            steps_per_epoch, dtype=np.float32
        )

        episode_returns_list: List[float] = []
        episode_lengths_list: List[int] = []

        # Variables on the current episode
        rewards: List[float] = []
        values: List[float] = []
        episode_length: int = 0

        observation: np.ndarray = self.env.reset()

        for current_step in range(steps_per_epoch):
            observation_tensor: Tensor = torch.from_numpy(observation).float()

            observations_list.append(observation_tensor)

            with torch.no_grad():
                policy_dist: Distribution = self.policy(observation_tensor)
                value: Tensor = self.value_function(observation_tensor)

            values.append(value.detach().item())

            action: Tensor = policy_dist.sample()

            actions_list.append(action)

            action_ndarray = action.detach().numpy()
            reward: float
            episode_done: bool
            observation, reward, episode_done, _ = self.env.step(action_ndarray)

            rewards.append(reward)

            episode_length += 1

            self.current_total_steps += 1

            epoch_ended: bool = current_step == steps_per_epoch - 1

            if episode_done or epoch_ended:
                if epoch_ended and not (episode_done):
                    logger.warning(
                        "The trajectory cut off at {} steps on the current episode".format(
                            episode_length
                        )
                    )

                last_value_float: float
                if epoch_ended:
                    observation_tensor = torch.from_numpy(observation).float()

                    with torch.no_grad():
                        last_value: Tensor = self.value_function(observation_tensor)

                    last_value_float = last_value.detach().item()
                else:
                    last_value_float = 0.0
                values.append(last_value_float)
                rewards.append(last_value_float)

                episode_slice = slice(
                    current_step - episode_length + 1, current_step + 1
                )

                # Calculate advantage over an episode
                values_ndarray: np.ndarray = np.asarray(values)
                rewards_ndarray: np.ndarray = np.asarray(rewards)
                episode_advantage: np.ndarray = gae(
                    rewards_ndarray, self.gamma, values_ndarray, self.gae_lambda
                )

                advantages_ndarray[episode_slice] = episode_advantage

                # Calculate rewards-to-go over an episode, to be targets for the value function
                episode_discounted_return: np.ndarray = discount_cumulative_sum(
                    rewards, self.gamma
                )[:-1]
                discounted_returns_ndarray[episode_slice] = episode_discounted_return

                episode_true_return: float = np.sum(rewards).item()

                episode_returns_list.append(episode_true_return)
                episode_lengths_list.append(episode_length)

                if episode_done:
                    self.current_total_episodes += 1

                observation, episode_length = self.env.reset(), 0
                rewards, values = [], []

        one_epoch_experience["observations"] = torch.stack(observations_list)
        one_epoch_experience["actions"] = torch.stack(actions_list)
        one_epoch_experience["advantages"] = torch.from_numpy(advantages_ndarray)
        one_epoch_experience["discounted_returns"] = torch.from_numpy(
            discounted_returns_ndarray
        )
        one_epoch_experience["episode_returns"] = episode_returns_list
        one_epoch_experience["episode_lengths"] = episode_lengths_list

        return one_epoch_experience

    @abstractmethod
    def train(self, one_epoch_experience: OneEpochExperience) -> None:
        """
        Train the algorithm with the experience

        :param one_epoch_experience: (OneEpochExperience) Collected experience on one epoch.
        """
        raise NotImplementedError

    def predict(self, observation: np.ndarray) -> np.ndarray:
        """
        Get the action(s) from an observation which are sampled under the current policy.

        :param observation: (np.ndarray) The input observation
        :return: (np.ndarray) The action(s)
        """
        observation_tensor: Tensor = torch.from_numpy(observation).float()
        action: Tensor = self.policy.predict(observation_tensor)
        action_ndarray: np.ndarray = action.detach().numpy()

        return action_ndarray



from abc import ABC, abstractmethod

from torch import Tensor, nn
from torch.optim import Optimizer


class Policy(nn.Module, ABC):
    """
    The base policy class

    :param network: (nn.Module) The network.
    :param optimizer: (Optimizer) The optimizer.
    """

    def __init__(self, network: nn.Module, optimizer: Optimizer):
        super().__init__()

        self.network = network
        self.optimizer = optimizer

    @abstractmethod
    def predict(self, observation: Tensor) -> Tensor:
        """
        Selects the action(s) based on the observation of the environment.

        :param observation: (Tensor) The observation(s) of the environment
        :return: (Tensor) the action(s)
        """
        raise NotImplementedError



import logging
from typing import Dict, List, Optional

import gym
import numpy as np
import torch
from torch import Tensor
from torch.nn import functional as F

from rl_replicas.common.base_algorithms import OffPolicyAlgorithm
from rl_replicas.common.policies import Policy
from rl_replicas.common.q_function import QFunction
from rl_replicas.common.replay_buffer import ReplayBuffer
from rl_replicas.common.utils import polyak_average

logger = logging.getLogger(__name__)


class DDPG(OffPolicyAlgorithm):
    """
    Deep Deterministic Policy Gradient (DDPG)

    :param policy: (Policy) The policy
    :param q_function: (QFunction) The Q function
    :param env: (gym.Env) The environment to learn from
    :param gamma: (float) The discount factor for the cumulative return
    :param tau: (float) The interpolation factor in polyak averaging for target networks
    :param action_noise_scale: (float) The scale of the noise (std)
    :param seed: (int) The seed for the pseudo-random generators
    """

    def __init__(
        self,
        policy: Policy,
        q_function: QFunction,
        env: gym.Env,
        gamma: float = 0.99,
        tau: float = 0.005,
        action_noise_scale: float = 0.1,
        seed: Optional[int] = None,
    ) -> None:
        super().__init__(
            policy=policy,
            q_function=q_function,
            env=env,
            gamma=gamma,
            tau=tau,
            action_noise_scale=action_noise_scale,
            seed=seed,
        )

    def train(
        self,
        replay_buffer: ReplayBuffer,
        train_steps: int,
        minibatch_size: int,
    ) -> None:
        policy_losses: List[float] = []
        q_function_losses: List[float] = []
        all_q_values: List[float] = []

        for _ in range(train_steps):
            minibatch: Dict[str, Tensor] = replay_buffer.sample_minibatch(
                minibatch_size
            )

        observations: Tensor = minibatch["observations"]
        actions: Tensor = minibatch["actions"]
        rewards: Tensor = minibatch["rewards"]
        next_observations: Tensor = minibatch["next_observations"]
        dones: Tensor = minibatch["dones"]

        q_values: Tensor = self.q_function(observations, actions)
        all_q_values.extend(q_values.tolist())

        with torch.no_grad():
            next_actions: Tensor = self.target_policy(next_observations)
            target_q_values: Tensor = self.target_q_function(
                next_observations, next_actions
            )

            targets: Tensor = rewards + self.gamma * (1 - dones) * target_q_values

        q_function_loss: Tensor = F.mse_loss(q_values, targets)
        q_function_losses.append(q_function_loss.item())

        self.q_function.optimizer.zero_grad()
        q_function_loss.backward()
        self.q_function.optimizer.step()

        policy_actions: Tensor = self.policy(observations)

        # Freeze Q-network so you don't waste computational effort
        for param in self.q_function.network.parameters():
            param.requires_grad = False

        policy_q_values: Tensor = self.q_function(observations, policy_actions)

        policy_loss: Tensor = -torch.mean(policy_q_values)
        policy_losses.append(policy_loss.item())

        self.policy.optimizer.zero_grad()
        policy_loss.backward()
        self.policy.optimizer.step()

        # Unfreeze Q-network
        for param in self.q_function.network.parameters():
            param.requires_grad = True

        polyak_average(
            self.policy.network.parameters(),
            self.target_policy.network.parameters(),
            self.tau,
        )
        polyak_average(
            self.q_function.network.parameters(),
            self.target_q_function.network.parameters(),
            self.tau,
        )

        if self.tensorboard:
            self.writer.add_scalar(
                "policy/loss",
                policy_loss,
                self.current_total_steps,
            )
            self.writer.add_scalar(
                "q-function/loss",
                q_function_loss,
                self.current_total_steps,
            )
            self.writer.add_scalar(
                "q-function/avarage_q-value",
                torch.mean(q_values),
                self.current_total_steps,
            )

        logger.info("Policy Loss:            {:<8.3g}".format(np.mean(policy_losses)))
        logger.info(
            "Q Function Loss:        {:<8.3g}".format(np.mean(q_function_losses))
        )

        logger.info("Average Q Value:        {:<8.3g}".format(np.mean(all_q_values)))
        logger.info("Max Q Value:            {:<8.3g}".format(np.max(all_q_values)))
        logger.info("Min Q Value:            {:<8.3g}".format(np.min(all_q_values)))



import copy
import logging
import os
import time
from abc import ABC, abstractmethod
from typing import List, Optional, Tuple

import gym
import numpy as np
import torch
from torch import Tensor
from torch.utils.tensorboard import SummaryWriter

from rl_replicas.common.policies import Policy
from rl_replicas.common.q_function import QFunction
from rl_replicas.common.replay_buffer import ReplayBuffer
from rl_replicas.common.utils import seed_random_generators

logger = logging.getLogger(__name__)


class OffPolicyAlgorithm(ABC):
    """
    The base of off-policy algorithms

    :param policy: (Policy) The policy
    :param q_function: (QFunction) The Q function
    :param env: (gym.Env) The environment to learn from
    :param gamma: (float) The discount factor for the cumulative return
    :param tau: (float) The interpolation factor in polyak averaging for target networks
    :param action_noise_scale: (float) The scale of the action noise (std)
    :param seed: (int) The seed for the pseudo-random generators
    """

    def __init__(
        self,
        policy: Policy,
        q_function: QFunction,
        env: gym.Env,
        gamma: float,
        tau: float,
        action_noise_scale: float,
        seed: Optional[int],
    ) -> None:
        self.policy = policy
        self.q_function = q_function
        self.env = env
        self.gamma = gamma
        self.tau = tau
        self.action_noise_scale = action_noise_scale
        if seed is not None:
            self.seed: int = seed

        self.evaluation_env = gym.make(env.spec.id)
        self.action_limit: float = self.env.action_space.high[0]
        self.action_size: int = self.env.action_space.shape[0]
        self.target_policy = copy.deepcopy(self.policy)
        self.target_q_function = copy.deepcopy(self.q_function)
        if seed is not None:
            self._seed()

        for param in self.target_policy.network.parameters():
            param.requires_grad = False
        for param in self.target_q_function.network.parameters():
            param.requires_grad = False

    def _seed(self) -> None:
        seed_random_generators(self.seed)
        self.env.action_space.seed(self.seed)
        self.env.seed(self.seed)

    def learn(
        self,
        epochs: int = 2000,
        steps_per_epoch: int = 50,
        replay_buffer_size: int = int(1e6),
        minibatch_size: int = 100,
        random_start_steps: int = 10000,
        steps_before_update: int = 1000,
        train_steps: int = 50,
        num_evaluation_episodes: int = 5,
        evaluation_interval: int = 4000,
        output_dir: str = ".",
        tensorboard: bool = False,
        model_saving: bool = False,
    ) -> None:
        """
        Learn the model

        :param epochs: (int) The number of epochs to run and train.
        :param steps_per_epoch: (int) The number of steps to run per epoch; in other words, batch size is
            steps_per_epoch.
        :param replay_size: (int) The size of the replay buffer
        ;param minibatch_size: (int) The minibatch size for SGD.
        :param random_start_steps: (int) The number of steps for uniform-random action selection for exploration
            at the beginning.
        :param steps_before_update: (int) The number of steps to perform before policy is updated.
        :param train_steps: (int) The number of training steps on each epoch
        :param num_evaluation_episodes: (int) The number of evaluation episodes
        :param evaluation_interval: (int) The interval steps of evaluation
        :param output_dir: (str) The directory of output
        :param tensorboard: (bool) Whether or not to log for tensorboard
        :param model_saving: (bool) Whether or not to save trained model (Save and overwrite at each end of epoch)
        """
        self.tensorboard = tensorboard

        start_time: float = time.time()
        self.current_total_steps: int = 0
        self.current_total_episodes: int = 0

        if self.tensorboard:
            logger.info("Set up tensorboard")
            os.makedirs(output_dir, exist_ok=True)
            tensorboard_path: str = os.path.join(output_dir, "tensorboard")
            self.writer: SummaryWriter = SummaryWriter(tensorboard_path)

        self.replay_buffer: ReplayBuffer = ReplayBuffer(replay_buffer_size)

        for current_epoch in range(epochs):
            episode_returns: List[float]
            episode_lengths: List[int]
            episode_returns, episode_lengths = self.collect_one_epoch_experience(
                self.replay_buffer, steps_per_epoch, random_start_steps
            )

            if model_saving:
                logger.info("Set up model saving")
                os.makedirs(output_dir, exist_ok=True)
                model_path: str = os.path.join(output_dir, "model.pt")

                logger.info("Save model")
                self.save_model(current_epoch, model_path)

            logger.info("Epoch: {}".format(current_epoch))

            logger.info(
                "Total steps:            {:<8.3g}".format(self.current_total_steps)
            )
            logger.info(
                "Total episodes:         {:<8.3g}".format(self.current_total_episodes)
            )

            if len(episode_lengths) > 0:
                logger.info(
                    "Average Episode Return: {:<8.3g}".format(np.mean(episode_returns))
                )
                logger.info(
                    "Episode Return STD:     {:<8.3g}".format(np.std(episode_returns))
                )
                logger.info(
                    "Max Episode Return:     {:<8.3g}".format(np.max(episode_returns))
                )
                logger.info(
                    "Min Episode Return:     {:<8.3g}".format(np.min(episode_returns))
                )

                logger.info(
                    "Average Episode Length: {:<8.3g}".format(np.mean(episode_lengths))
                )

                if self.tensorboard:
                    self.writer.add_scalar(
                        "training/average_episode_return",
                        np.mean(episode_returns),
                        self.current_total_steps,
                    )
                    self.writer.add_scalar(
                        "training/average_episode_length",
                        np.mean(episode_lengths),
                        self.current_total_steps,
                    )

            if (
                num_evaluation_episodes > 0
                and self.current_total_steps % evaluation_interval == 0
            ):
                self.evaluate_policy(num_evaluation_episodes, self.evaluation_env)

            logger.info(
                "Time:                   {:<8.3g}".format(time.time() - start_time)
            )

            if self.current_total_steps >= steps_before_update:
                self.train(self.replay_buffer, train_steps, minibatch_size)

        if self.tensorboard:
            self.writer.flush()
            self.writer.close()

    def collect_one_epoch_experience(
        self, replay_buffer: ReplayBuffer, steps_per_epoch: int, random_start_steps: int
    ) -> Tuple[List[float], List[int]]:
        observations_list: List[Tensor] = []
        actions_list: List[Tensor] = []
        next_observations_list: List[Tensor] = []

        rewards: List[float] = []
        dones: List[bool] = []

        episode_returns: List[float] = []
        episode_lengths: List[int] = []

        if not hasattr(self, "observation"):
            # Variables on the current episode
            self.episode_length: int = 0
            self.episode_return: float = 0

            self.observation: np.ndarray = self.env.reset()

        for current_step in range(steps_per_epoch):
            observation_tensor: Tensor = torch.from_numpy(self.observation).float()
            observations_list.append(observation_tensor)

            action: np.ndarray
            if self.current_total_steps < random_start_steps:
                action = self.env.action_space.sample()
            else:
                action = self.action_limit * self.predict(self.observation)
                action += self.action_noise_scale * np.random.randn(self.action_size)
                action = np.clip(action, -self.action_limit, self.action_limit)

            action_tensor: Tensor = torch.from_numpy(action).float()
            actions_list.append(action_tensor)

            next_observation: np.ndarray
            reward: float
            episode_done: bool
            next_observation, reward, episode_done, _ = self.env.step(action)

            next_observation_tensor: Tensor = torch.from_numpy(next_observation).float()
            next_observations_list.append(next_observation_tensor)

            self.observation = next_observation

            rewards.append(reward)
            dones.append(episode_done)

            self.current_total_steps += 1
            self.episode_length += 1
            self.episode_return += reward

            if episode_done:
                self.current_total_episodes += 1

                episode_returns.append(self.episode_return)
                episode_lengths.append(self.episode_length)

                self.observation, self.episode_length, self.episode_return = (
                    self.env.reset(),
                    0,
                    0,
                )

        replay_buffer.add_one_epoch_experience(
            observations_list, actions_list, rewards, next_observations_list, dones
        )

        return episode_returns, episode_lengths

    @abstractmethod
    def train(
        self, replay_buffer: ReplayBuffer, train_steps: int, minibatch_size: int
    ) -> None:
        """
        Train the algorithm with the experience.

        :param replay_buffer: (ReplayBuffer) The reply buffer
        :param train_steps: (int) The number of gradient descent updates
        :param minibatch_size: (int) The minibatch size
        """
        raise NotImplementedError

    def predict(self, observation: np.ndarray) -> np.ndarray:
        """
        Get the action(s) from an observation which are sampled under the current policy.

        :param observation: (np.ndarray) The input observation
        :return: (np.ndarray) The action(s)
        """
        observation_tensor: Tensor = torch.from_numpy(observation).float()
        action: Tensor = self.policy.predict(observation_tensor)
        action_ndarray: np.ndarray = action.detach().numpy()

        return action_ndarray

    def evaluate_policy(
        self,
        num_evaluation_episodes: int,
        evaluation_env: gym.Env,
    ) -> None:
        episode_returns: List[float] = []
        episode_lengths: List[int] = []

        for _ in range(num_evaluation_episodes):
            observation: np.ndarray = evaluation_env.reset()
            done: bool = False
            episode_return: float = 0.0
            episode_length: int = 0

            while not done:
                action: np.ndarray = self.predict(observation)

                reward: float
                observation, reward, done, _ = evaluation_env.step(action)

                episode_return += reward
                episode_length += 1

            episode_returns.append(episode_return)
            episode_lengths.append(episode_length)

        logger.info(
            "Average Evaluation Episode Return: {:<8.3g}".format(
                np.mean(episode_returns)
            )
        )
        logger.info(
            "Evaluation Episode Return STD:     {:<8.3g}".format(
                np.std(episode_returns)
            )
        )
        logger.info(
            "Max Evaluation Episode Return:     {:<8.3g}".format(
                np.max(episode_returns)
            )
        )
        logger.info(
            "Min Evaluation Episode Return:     {:<8.3g}".format(
                np.min(episode_returns)
            )
        )

        logger.info(
            "Average Evaluation Episode Length: {:<8.3g}".format(
                np.mean(episode_lengths)
            )
        )

        if self.tensorboard:
            self.writer.add_scalar(
                "evaluation/average_episode_return",
                np.mean(episode_returns),
                self.current_total_steps,
            )
            self.writer.add_scalar(
                "evaluation/average_episode_length",
                np.mean(episode_lengths),
                self.current_total_steps,
            )

    def save_model(self, current_epoch: int, model_path: str) -> None:
        """
        Save model

        :param current_epoch: (int) The current epoch
        :param model_path: (int) The path to save the model
        """
        torch.save(
            {
                "epoch": current_epoch,
                "total_steps": self.current_total_steps,
                "policy_state_dict": self.policy.network.state_dict(),
                "policy_optimizer_state_dict": self.policy.optimizer.state_dict(),
                "target_policy_state_dict": self.target_policy.network.state_dict(),
                "q_function_state_dict": self.q_function.network.state_dict(),
                "q_function_optimizer_state_dict": self.q_function.optimizer.state_dict(),
                "target_q_function_state_dict": self.target_q_function.network.state_dict(),
            },
            model_path,
        )



import torch
from torch import Tensor, nn
from torch.optim import Optimizer

from rl_replicas.common.policies.policy import Policy


class DeterministicPolicy(Policy):
    """
    The deterministic policy

    :param network: (nn.Module) The network.
    :param optimizer: (Optimizer) The optimizer.
    """

    def __init__(self, network: nn.Module, optimizer: Optimizer):
        super().__init__(network, optimizer)

    def forward(self, observation: Tensor) -> Tensor:
        """
        Forward pass in policy

        :param observation: (Tensor) The observation of the environment
        :return: (Tensor) The action(s).
        """
        action: Tensor = self.network(observation)

        return action

    def predict(self, observation: Tensor) -> Tensor:
        """
        Selects the action(s) based on the observation of the environment.

        :param observation: (Tensor) The observation(s) of the environment
        :return: (Tensor) the action(s)
        """
        with torch.no_grad():
            action: Tensor = self.forward(observation)

        return action



import logging
from typing import Optional

import gym
import torch
from torch import Tensor
from torch.distributions import Distribution
from torch.nn import functional as F

from rl_replicas.common.base_algorithms.on_policy_algorithm import (
    OneEpochExperience,
    OnPolicyAlgorithm,
)
from rl_replicas.common.policies import Policy
from rl_replicas.common.value_function import ValueFunction

logger = logging.getLogger(__name__)


class VPG(OnPolicyAlgorithm):
    """
    Vanilla Policy Gradient (REINFORCE) with GAE for advantage estimation

    :param policy: (Policy) The policy
    :param value_function: (ValueFunction) The value function
    :param env: (gym.Env) The environment to learn from
    :param gamma: (float) The discount factor for the cumulative return
    :param gae_lambda: (float) The factor for trade-off of bias vs variance for Generalized Advantage Estimator
    :param seed: (int) The seed for the pseudo-random generators
    :param n_value_gradients (int): Number of gradient descent steps to take on value function per epoch.
    """

    def __init__(
        self,
        policy: Policy,
        value_function: ValueFunction,
        env: gym.Env,
        gamma: float = 0.99,
        gae_lambda: float = 0.97,
        seed: Optional[int] = None,
        n_value_gradients: int = 80,
    ) -> None:
        super().__init__(
            policy=policy,
            value_function=value_function,
            env=env,
            gamma=gamma,
            gae_lambda=gae_lambda,
            seed=seed,
            n_value_gradients=n_value_gradients,
        )

    def train(self, one_epoch_experience: OneEpochExperience) -> None:
        observations: Tensor = one_epoch_experience["observations"]
        actions: Tensor = one_epoch_experience["actions"]
        advantages: Tensor = one_epoch_experience["advantages"]

        # Normalize advantage
        advantages = (advantages - torch.mean(advantages)) / torch.std(advantages)

        policy_dist: Distribution = self.policy(observations)
        log_probs: Tensor = policy_dist.log_prob(actions)

        policy_loss: Tensor = -torch.mean(log_probs * advantages)

        # for logging
        policy_loss_before: Tensor = policy_loss.detach()
        entropies: Tensor = policy_dist.entropy().detach()

        # Train policy
        self.policy.optimizer.zero_grad()
        policy_loss.backward()
        self.policy.optimizer.step()

        discounted_returns: Tensor = one_epoch_experience["discounted_returns"]

        # for logging
        with torch.no_grad():
            value_loss_before: Tensor = self.compute_value_loss(
                observations, discounted_returns
            )

        # Train value function
        for _ in range(self.n_value_gradients):
            value_loss: Tensor = self.compute_value_loss(
                observations, discounted_returns
            )
            self.value_function.optimizer.zero_grad()
            value_loss.backward()
            self.value_function.optimizer.step()

        logger.info("Policy Loss:            {:<8.3g}".format(policy_loss_before))
        logger.info("Avarage Entropy:        {:<8.3g}".format(torch.mean(entropies)))
        logger.info("Log Prob STD:           {:<8.3g}".format(torch.std(log_probs)))

        logger.info("Value Function Loss:    {:<8.3g}".format(value_loss_before))

        if self.tensorboard:
            self.writer.add_scalar(
                "policy/loss", policy_loss_before, self.current_total_steps
            )
            self.writer.add_scalar(
                "policy/avarage_entropy",
                torch.mean(entropies),
                self.current_total_steps,
            )
            self.writer.add_scalar(
                "policy/log_prob_std", torch.std(log_probs), self.current_total_steps
            )

            self.writer.add_scalar(
                "value/loss", value_loss_before, self.current_total_steps
            )

    def compute_value_loss(
        self, observations: Tensor, discounted_returns: Tensor
    ) -> Tensor:
        values: Tensor = self.value_function(observations)
        squeezed_values: Tensor = torch.squeeze(values, -1)
        value_loss: Tensor = F.mse_loss(squeezed_values, discounted_returns)

        return value_loss



import random
from typing import Iterable, List

import numpy as np
import scipy.signal
import torch
from torch import Tensor


def discount_cumulative_sum(vector: np.ndarray, discount: float) -> np.ndarray:
    """
    Compute discounted cumulative sums of vector.

    :param vector: (np.ndarray) An target vector
    e.g. [x0,
          x1,
          x2]
    :param discount: (float) The discount factor for the cumulative return
    :return: (np.ndarray) discounted cumulative sums of an vector
    e.g. [x0 + discount * x1 + discount^2 * x2,
          x1 + discount * x2,
          x2]
    """
    return scipy.signal.lfilter([1], [1, -discount], vector[::-1], axis=0)[::-1]


def seed_random_generators(seed: int) -> None:
    """
    Set the seed of the pseudo-random generators (python, numpy, pytorch)

    :param seed: (int)
    """
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)


def gae(
    rewards: np.ndarray, gamma: float, values: np.ndarray, gae_lambda: float
) -> np.ndarray:
    """
    Compute Generalized Advantage Estimation (GAE)

    :param rewards: (np.ndarray) Rewards for all states
    :param gamma: (float) The discount factor for the cumulative return
    :param values: (np.ndarray) Values for all states
    :param gae_lambda: (float) A smoothing parameter for reducing the variance
    :return gaes: (np.ndarray) GAEs for all states
    """
    deltas: np.ndarray = rewards[:-1] + gamma * values[1:] - values[:-1]
    gaes: np.ndarray = discount_cumulative_sum(deltas, gamma * gae_lambda)

    return gaes


def unflatten_tensors(
    flattened: np.ndarray, tensor_shapes: List[torch.Size]
) -> List[Tensor]:
    """
    Unflatten a flattened tensors into a list of tensors.

    :param flattened: (np.ndarray) Flattened tensors.
    :param tensor_shapes: (List[torch.Size]) Tensor shapes.
    :return: (List[np.ndarray]) Unflattened list of tensors.
    """
    tensor_sizes = list(map(np.prod, tensor_shapes))
    indices = np.cumsum(tensor_sizes)[:-1]

    return [
        np.reshape(pair[0], pair[1])
        for pair in zip(np.split(flattened, indices), tensor_shapes)
    ]


def polyak_average(
    params: Iterable[torch.nn.Parameter],
    target_params: Iterable[torch.nn.Parameter],
    tau: float,
) -> None:
    """
    Perform Polyak averaging on target_params using params.

    :param params: (Iterable[torch.nn.Parameter]) The parameters to use to update the target params
    :param target_params: (Iterable[torch.nn.Parameter]) The parameters to update
    :param tau: (float) the soft update coefficient ("Polyak update", between 0 and 1)
    """
    with torch.no_grad():
        for param, target_param in zip(params, target_params):
            target_param.data.copy_((1.0 - tau) * target_param.data + tau * param.data)
